@@ -17,6 +17,10 @@ use std::{
     time::{Duration, SystemTime, UNIX_EPOCH},
 };
 
+// When querying Docker for the image IDs corresponding to a list of container IDs, this is the
+// maximum number of container IDs to query at once.
+const CONTAINER_IDS_CHUNK_SIZE: usize = 100;
+
 // A Docker event (a line of output from `docker events --format '{{json .}}'`)
 #[derive(Deserialize, Serialize, Debug)]
 struct Event {
@@ -105,11 +109,16 @@ fn image_id(image: &str) -> io::Result<String> {
         .map_err(|error| io::Error::new(io::ErrorKind::Other, error))
 }
 
-// Ask Docker for the ID of the parent of an image.
-fn parent_id(image: &str) -> io::Result<Option<String>> {
+// Get the ID of the parent of an image (if the parent exists), querying Docker if necessary.
+fn parent_id(state: &State, image_id: &str) -> io::Result<Option<String>> {
+    // If we already know the parent, just return it.
+    if let Some(image) = state.images.get(image_id) {
+        return Ok(image.parent_id.clone());
+    }
+
     // Query Docker for the parent image ID.
     let output = Command::new("docker")
-        .args(&["image", "inspect", "--format", "{{.Parent}}", image])
+        .args(&["image", "inspect", "--format", "{{.Parent}}", image_id])
         .stderr(Stdio::inherit())
         .output()?;
 
@@ -119,7 +128,7 @@ fn parent_id(image: &str) -> io::Result<Option<String>> {
             io::ErrorKind::Other,
             format!(
                 "Unable to determine ID of the parent of image {}.",
-                image.code_str(),
+                image_id.code_str(),
             ),
         ));
     }
@@ -187,11 +196,7 @@ fn list_image_records(state: &mut State) -> io::Result<HashMap<String, ImageReco
                 }
                 Entry::Vacant(entry) => {
                     entry.insert(ImageRecord {
-                        parent_id: if let Some(image) = state.images.get(id) {
-                            image.parent_id.clone()
-                        } else {
-                            parent_id(id)?
-                        },
+                        parent_id: parent_id(state, id)?,
                         created_since_epoch: parse_docker_date(date_str)?,
                         repository_tags: vec![repository_tag],
                     });
@@ -210,51 +215,88 @@ fn list_image_records(state: &mut State) -> io::Result<HashMap<String, ImageReco
 
 // Ask Docker for the IDs of the images currently in use by containers.
 fn image_ids_in_use() -> io::Result<HashSet<String>> {
-    // Query Docker for the image IDs.
-    let output = Command::new("docker")
+    // Query Docker for the container IDs.
+    let container_ids_output = Command::new("docker")
         .args(&[
             "container",
             "ls",
             "--all",
             "--no-trunc",
             "--format",
-            "{{.Image}}",
+            "{{.ID}}",
         ])
         .stderr(Stdio::inherit())
         .output()?;
 
     // Ensure the command succeeded.
-    if !output.status.success() {
+    if !container_ids_output.status.success() {
         return Err(io::Error::new(
             io::ErrorKind::Other,
             "Unable to determine IDs of images currently in use by containers.",
         ));
     }
 
-    // Interpret the output bytes as UTF-8 and collect the lines.
-    String::from_utf8(output.stdout)
+    // Interpret the output bytes as UTF-8 and parse the lines.
+    let container_ids = String::from_utf8(container_ids_output.stdout)
         .map_err(|error| io::Error::new(io::ErrorKind::Other, error))
         .map(|output| {
             output
                 .lines()
                 .filter_map(|line| {
-                    if line.is_empty() {
+                    let trimmed_line = line.trim();
+
+                    if trimmed_line.is_empty() {
                         None
                     } else {
-                        match image_id(line.trim()) {
-                            Ok(the_image_id) => Some(the_image_id),
-                            Err(error) => {
-                                // This failure may happen if a container was created from an image
-                                // that no longer exists. This is non-fatal, so we just log the
-                                // error and continue.
-                                error!("{}", error);
-                                None
-                            }
-                        }
+                        Some(trimmed_line.to_owned())
                     }
                 })
-                .collect()
-        })
+                .collect::<Vec<_>>()
+        })?;
+
+    // Group the container IDs into chunks and query Docker for the image IDs for each chunk.
+    let mut image_ids = HashSet::new();
+    for chunk in container_ids.chunks(CONTAINER_IDS_CHUNK_SIZE) {
+        // Query Docker for the image IDs for this chunk.
+        let image_ids_output = Command::new("docker")
+            .args(
+                vec!["container", "inspect", "--format", "{{.Image}}"]
+                    .into_iter()
+                    .chain(chunk.iter().map(AsRef::as_ref)),
+            )
+            .stderr(Stdio::inherit())
+            .output()?;
+
+        // Ensure the command succeeded.
+        if !image_ids_output.status.success() {
+            return Err(io::Error::new(
+                io::ErrorKind::Other,
+                "Unable to determine IDs of images currently in use by containers.",
+            ));
+        }
+
+        // Interpret the output bytes as UTF-8 and parse the lines.
+        image_ids.extend(
+            String::from_utf8(image_ids_output.stdout)
+                .map_err(|error| io::Error::new(io::ErrorKind::Other, error))
+                .map(|output| {
+                    output
+                        .lines()
+                        .filter_map(|line| {
+                            let trimmed_line = line.trim();
+
+                            if trimmed_line.is_empty() {
+                                None
+                            } else {
+                                Some(trimmed_line.to_owned())
+                            }
+                        })
+                        .collect::<Vec<_>>()
+                })?,
+        );
+    }
+
+    Ok(image_ids)
 }
 
 // Get the total space used by Docker images.
@@ -342,13 +384,6 @@ fn touch_image(state: &mut State, image_id: &str, verbose: bool) -> io::Result<(
         );
     }
 
-    // Find the parent of the image, either by looking it up in the state or by querying Docker.
-    let parent_id = if let Some(image) = state.images.get(image_id) {
-        image.parent_id.clone()
-    } else {
-        parent_id(image_id)?
-    };
-
     // Get the current timestamp.
     match SystemTime::now().duration_since(UNIX_EPOCH) {
         Ok(duration) => {
@@ -356,7 +391,7 @@ fn touch_image(state: &mut State, image_id: &str, verbose: bool) -> io::Result<(
             state.images.insert(
                 image_id.to_owned(),
                 state::Image {
-                    parent_id,
+                    parent_id: parent_id(state, image_id)?,
                     last_used_since_epoch: duration,
                 },
             );
